#include <Engine/Globals.h>
#include <Engine/types/Scene.h>
#include <Engine/types/entity/Camera.h>
#include <Engine/types/entity/InstancedSprite.h>
#include <MultiThreading/Thread.h>
#include <Engine/Logger.h>
#include <JGL/JGL.h>
#include <algorithm>
#include <thread>

using namespace Engine;
size_t Scene::EntityCount() const {
    return Entity::GetAllDescendents(entity_list).size();
}

void Scene::Update() {
    entity_list->UpdateChildren();
}

void Scene::UpdateLoop() {
    while (!destroying) {
        if (!should_tick) {
            auto target_tick_time = std::chrono::microseconds(1000000 / tick_rate);
            std::this_thread::sleep_for(std::chrono::microseconds(target_tick_time));
            continue;
        }
        auto target_tick_time = std::chrono::microseconds(1000000 / tick_rate);
        auto start = std::chrono::high_resolution_clock::now();

        Update();

        auto stop = std::chrono::high_resolution_clock::now();
        auto elapsed = std::chrono::duration_cast<std::chrono::microseconds>(stop - start);

        if (elapsed < target_tick_time)
            std::this_thread::sleep_for(std::chrono::microseconds(target_tick_time - elapsed));

        // TODO Research how we can fix desync issues that would come about by a tick occasionally taking longer than we expected.
        if (elapsed > target_tick_time)
            Logger::Info("This tick took " + std::to_string((elapsed - target_tick_time).count()) + "μs longer than the target tick time.");

        last_tick_time = std::chrono::duration_cast<std::chrono::microseconds>(stop.time_since_epoch()).count();
    }
}

void Scene::Render() {
    std::vector<Renderable*> display_list{};
    for (auto *e: Entity::GetAllDescendents(entity_list))
        if (auto *r = dynamic_cast<Renderable *>(e))
            if (r->ShouldRender())
                display_list.push_back(r);

    std::sort(display_list.begin(), display_list.end(), [](Renderable *a, Renderable *b) {
        return a->GetLayer() > b->GetLayer();
    });

    if (active_camera)
        active_camera->Render();

    for (auto &r: display_list)
        if (r->ShouldRender())
            r->Render();
    J2D::End();
}

void Scene::Render(JGL::RenderTarget* render_target) {
    auto start = std::chrono::steady_clock::now();
    if (!render_target)
        Logger::Fatal("Rendering the current scene to a RenderTarget that is nullptr?");

    J2D::Begin(render_target, true);
        Render();
    J2D::End();
    auto stop = std::chrono::steady_clock::now();

    auto frame_time = std::chrono::duration_cast<std::chrono::microseconds>(stop - start).count();
    render_target_frame_delta = frame_time / (1000.f * 1000.f);
    last_rendered_to_window = false;
}

void Scene::Render(ReWindow::RWindow* window) {
    if (!window)
        Logger::Fatal("Rendering the current scene to a ReWindow that is nullptr?");

    J2D::Begin(nullptr, true);
        Render();
    J2D::End();
    last_rendered_to_window = true;
}

Scene::~Scene() {
    Scene::Unload();
}

void Scene::Unload() {
    destroying = true;

    for (auto& itx : instanced_textures)
        delete itx.second;

    delete update_worker;
    delete entity_list;
    delete active_camera;
}

std::string Scene::GetName() const {
    return name;
}

Camera* Scene::GetActiveCamera() const {
    return active_camera;
}

Texture* Scene::GetInstancedTexture(const InstancedSprite* user) {
    auto it = instanced_textures.find(typeid(user).hash_code());
    if (it != instanced_textures.end())
        return it->second;

    auto* t = new Texture(user->GetTextureFilesystemPath());
    instanced_textures.insert(std::make_pair(typeid(user).hash_code(), t));
    return t;
}

void Scene::DestroyInstancedTexture(const InstancedSprite* user) {
    auto it = instanced_textures.find(typeid(user).hash_code());
    if (it != instanced_textures.end()) {
        delete it->second;
        instanced_textures.erase(it);
    }
}

void Scene::Run() {
    update_worker = new MultiThreading::Thread([this] { UpdateLoop(); } );
}

bool Scene::ShouldTick() const {
    return should_tick;
}

void Scene::ShouldTick(bool state) {
    should_tick = state;
}

float Scene::TickDeltaTime() const {
    return (1000000.f / tick_rate) / 1000000.f;
}

float Scene::FrameDeltaTime() const {
    if (last_rendered_to_window && Globals::Window)
        return Globals::Window->GetDeltaTime();
    return render_target_frame_delta;
}

unsigned long Scene::LastTickTime() const {
    return last_tick_time;
}

u8 Scene::GetTickRate() const {
    return tick_rate;
}

