#include <Engine/types/entity/mixins/Movable.h>
#include <Engine/Globals.h>

using namespace J3ML;
using namespace Engine;

float Movable::NormalizeRadians(float radians) const {
    radians = std::fmod(radians, Math::Pi * 2);
    if (radians < 0)
        radians += Math::Pi * 2;
    return radians;
}

void Movable::SetHeading(float new_heading) {
    heading = NormalizeRadians(new_heading);
}

float Movable::GetHeading() const {
    return heading;
}

Vector2 Movable::GetPosition() const {
    return position;
}

float Movable::GetDecay() const {
    return decay;
}

float Movable::GetVelocity() const {
    return velocity;
}

float Movable::GetFriction() const {
    return friction;
}

void Movable::SetDecay(float new_decay) {
    decay = std::clamp(new_decay, 0.0f, 1.0f);
}

void Movable::SetFriction(float new_friction) {
    friction = std::clamp(new_friction, 0.0f, 1.0f);
}

void Movable::SetVelocity(float new_velocity) {
    velocity = new_velocity;
}

void Movable::SetPosition(const Vector2& new_position, bool reset) {
    position = new_position;
    if (reset) {
        heading = 0;
        velocity = 0;
        friction = 0;
        decay = 0;
    }
}

void Movable::SetConstantForceHeading(float heading_rad) {
    constant_force_heading = heading_rad;
}

void Movable::SetConstantForce(float new_velocity) {
    constant_force_velocity = new_velocity;
}

float Movable::GetConstantForceHeading() const {
    return constant_force_heading;
}

float Movable::GetConstantForceVelocity() const {
    return constant_force_velocity;
}

void Movable::SetVisualRotation(float rad_rotation) {
    face_angle = rad_rotation;
}

float Movable::GetVisualRotation() const {
    return face_angle;
}

void Movable::Move(const u8& tick_rate) {
    previous_decay = decay;
    previous_velocity = velocity;
    previous_heading = heading;
    previous_face_angle = face_angle;
    previous_constant_force_heading = constant_force_heading;
    previous_constant_force_velocity = velocity;
    previous_friction = friction;
    previous_position = position;

    // Advance.
    position.x += ((Math::Cos(heading) * velocity) + (Math::Cos(constant_force_heading) * constant_force_velocity)) * (1.f / tick_rate);
    position.y += ((Math::Sin(heading) * velocity) + (Math::Sin(constant_force_heading) * constant_force_velocity))  * (1.f / tick_rate);
    
    // Apply decay.
    velocity = velocity * (1 - decay);
    // Apply friction.
    velocity = velocity * (1 - friction);
}

void Movable::AddVelocity(float heading_rad, float vel) {
    heading_rad = NormalizeRadians(heading_rad);

    Vector2 v = {velocity * Math::Cos(heading), velocity * Math::Sin(heading)};
    Vector2 add = {vel * Math::Cos(heading_rad), vel * Math::Sin(heading_rad)};
    Vector2 result = {v.x + add.x, v.y + add.y};

    velocity = Math::Sqrt(result.x * result.x + result.y * result.y);
    heading = NormalizeRadians(Math::Atan2(result.y, result.x));
}

float Movable::HeadingRelativeForward() const {
    return heading;
}

float Movable::FaceRelativeForward() const {
    return face_angle;
}

float Movable::FaceRelativeBackward() const {
    return NormalizeRadians(face_angle + Math::Pi);
}

float Movable::HeadingRelativeBackward() const {
    return NormalizeRadians(heading + Math::Pi);
}

float Movable::FaceRelativeLeft() const {
    return NormalizeRadians(face_angle - Math::HalfPi);
}

float Movable::HeadingRelativeLeft() const {
    return NormalizeRadians(heading - Math::HalfPi);
}

float Movable::FaceRelativeRight() const {
    return NormalizeRadians(face_angle + Math::HalfPi);
}

float Movable::HeadingRelativeRight() const {
    return NormalizeRadians(heading + Math::HalfPi);
}

float Movable::WorldUp() {
    return -Math::Pi / 2;
}

float Movable::WorldDown() {
    return Math::Pi / 2;
}

float Movable::WorldLeft() {
    return Math::Pi;
}

float Movable::WorldRight() {
    return 0;
}

